/*
    egg.inc - macros implementing Narushin's egg function
    Copyright (C) 2022  Pascal Baillehache info@baillehachepascal.dev
    https://baillehachepascal.dev
    This script is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.
    This script is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

// Top level macro creating an egg with given length and maximum
// radius, and random shape and texture
// Inputs:
//     lengthEgg: length of the egg from the tip of the large end to the
//                tip of the small end
//     radiusEgg: radius of the egg (from the rotation axis passing
//                through the tips) at its largest position
//     seedShape: seed for the PRNG used to create the shape
//   seedTexture: seed for the PRNG used to create the texture
// The tip of the big end is located at <0,0,0>. The tip of both ends
// lie on the y axis.
#macro EggRndShapeRndTexture(lengthEgg, radiusEgg, seedShape, seedTexture)
  object {
    EggRndShape(lengthEgg, radiusEgg, seedShape)
    EggRndTexture(lengthEgg, radiusEgg, seedTexture)
  }
#end

// Macro creating an egg with given length and maximum radius, and
// random shape
// Inputs:
//   lengthEgg: length of the egg from the tip of the large end to the
//              tip of the small end
//   radiusEgg: radius of the egg (from the rotation axis passing
//              through the tips) at its largest position
//   seedShape: seed for the PRNG used to create the shape
// The tip of the big end is located at <0,0,0>. The tip of both ends
// lie on the y axis.
#macro EggRndShape(lengthEgg, radiusEgg, seedShape)
  // Create PRNG for the shape
  #local rndSeedShape = seed(seedShape);
  // Calculate the parameters 'w' and 'D' of Narushin's fornmula.
  // Randomly chosen in a way to keep them realistic according to
  // the ratio radius/length. If it's equal to 1.0, it will produce a
  // spheric egg, else it will produce one of the three other types.
  #local t_ = radiusEgg / lengthEgg;
  #local w = lengthEgg / 8 * rand(rndSeedShape) * (1- t_);
  #local r = .75 + .25 * rand(rndSeedShape);
  #local D = radiusEgg * 2 * 0.866025 * (r + t_ * (1 - r));
  // Create the egg given the four parameters
  EggShape(lengthEgg, radiusEgg * 2, w, D)
#end

// Macro creating an egg given the four parameters of Narushin's formula
// Inputs:
//   L: length from tip of bug end to tip of small end
//   B: maximum breadth
//   w: distance from half length to position of the maximum breadth
//   D: breadth at L/4 from tip of small end
// The tip of the big end is located at <0,0,0>. The tip of both ends
// lie on the y axis.
#macro EggShape(L, B, w, D)
  // Uses a lathe with cubic_spline to create the egg as a surface of
  // revolution
  lathe {
    cubic_spline
    // Number of points on the lathe
    #local nbPoint = 100;
    // Uses two extra points for the cubic interpolation at the
    // extremities.
    (nbPoint+2),
    // The first extra point is the mirrored point of the second point.
    #local x_ = (1 / (nbPoint - 1) - .5) * L;
    <-1 * Narushin(x_), x_>
    // Loop on the points along the egg surface
    #local iPoint = 0;
    #while (iPoint < nbPoint)
      #local x_ = (iPoint / (nbPoint - 1) - .5) * L;
      <Narushin(x_), x_>
      #local iPoint = iPoint + 1;
    #end
    // The second extra point is the mirrored point of the penultieme
    // point.
    #local x_ = ((nbPoint - 2) / (nbPoint - 1) - .5) * L;
    <-1 * Narushin(x_), x_>
    // Translate the shape to position the tip of the big end at <0,0,0>
    translate L / 2 * y
  }
#end

// Narushin's formula
#macro Narushin(x_)
  #local y1 = B/2*sqrt((L*L-4*x_*x_)/(L*L+8*w*x_+4*w*w));
  #local y2 = sqrt(5.5*L*L+11*L*w+4*w*w)*(sqrt(3)*B*L-2*D*sqrt(L*L+2*w*L+4*w*w))/(sqrt(3)*B*L*(sqrt(5.5*L*L+11*L*w+4*w*w)-2*sqrt(L*L+2*w*L+4*w*w)));
  #local y3 = sqrt(L*(L*L+8*w*x_+4*w*w)/(2*(L-2*w)*x_*x_+(L*L+8*L*w-4*w*w)*x_+2*L*w*w+L*L*w+L*L*L));
  #local y_ = y1 * (1 - y2 * (1 - y3));
  y_
#end

// Macro creating the random texture of an egg with given length and
// maximum radius
// Inputs:
//     lengthEgg: length of the egg from the tip of the large end to the
//                tip of the small end
//     radiusEgg: radius of the egg (from the rotation axis passing
//                through the tips) at its largest position
//   seedTexture: seed for the PRNG used to create the texture
#macro EggRndTexture(lengthEgg, radiusEgg, seedTexture)
  // Create the PRNG for the texture
  #local rndSeedTexture = seed(seedTexture);
  material {
    // Main texture
    texture {
      // Main color, biased toward light tone
      #local r = 1 - pow(rand(rndSeedTexture), 2);
      #local g = 1 - pow(rand(rndSeedTexture), 2);
      #local b = 1 - pow(rand(rndSeedTexture), 2);
      // Variables to create a non uniform texture
      #local rd = .05 * rand(rndSeedTexture);
      #local gd = .05 * rand(rndSeedTexture);
      #local bd = .05 * rand(rndSeedTexture);
      pigment {
        // Uses a bozo pigment map with small deviation from the main
        // color to avoid a boring uniform texture
        bozo
        pigment_map {
          [0.0 color rgb <r-rd,g-gd,b-bd>]
          [0.5 color rgb <r,g,b>]
          [1.0 color rgb <r+rd,g+gd,b+bd>]
        }
        scale radiusEgg * 2 * rand(rndSeedTexture)
      }
      // Random normal bump
      #local bs = .05 + .25 * rand(rndSeedTexture);
      normal { bumps bs scale 0.5 * radiusEgg }
      // Random finish
      #local p = .1 + .3 * rand(rndSeedTexture);
      #local ps = 5 + 50 * rand(rndSeedTexture);
      finish { phong p phong_size ps}
      // Random translate to avoid uniformity
      translate rand(rndSeedTexture) * 10
    }
    // Add a random number of layered textures to create the stains
    #local iLayer = 0;
    #local nbLayer = 0;
    #while (iLayer < 3)
      // Randomise the existence of the layer at each level
      #if (rand(rndSeedTexture) < 0.33)
        texture {
          pigment {
            // Uses another bozo pigment map with parameters depending
            // on the layer level
            bozo
            #local map2 = .8 + .15 * rand(rndSeedTexture) - iLayer * .05;
            #local map1 = map2 - .4 * rand(rndSeedTexture) - iLayer * .05;
            #local r = rand(rndSeedTexture);
            #local g = rand(rndSeedTexture);
            #local b = rand(rndSeedTexture);
            color_map {
              [0.0 color rgbt <1, 1, 1, 1> ]
              [map1 color rgbt <1, 1, 1, 1> ]
              [map2 color rgbt <r,g,b,0> ]
              [1.0 color rgbt <r,g,b,0> ]
            }
            // Warp and random rotate to avoid uniformity
            warp { turbulence .5 }
            rotate <rand(rndSeedTexture), rand(rndSeedTexture), rand(rndSeedTexture)> * 360
          }
          // Parameterise also the size of the stains based on the layer
          // level
          scale radiusEgg * (1 + iLayer / 3)
        }
      #end
      #local iLayer = iLayer + 1;
    #end
    scale <radiusEgg, lengthEgg, radiusEgg>
  }
#end

